MAX_ENG_WORD_LENGTH = 45

# Trie Node Class
# value    - each node has a value which is a single character
# children - each node has a dictionary of child nodes. the key is the 
#            character represented by the child node
# is_word  - a boolean value indicating if this node is the end of
#            a "word path". 
class Node:
    def __init__(self, value, is_word):
        self.value = value
        self.children = {}
        self.is_word = is_word

def add_to_trie(word, root):
    # recursion base case
    if len(word) == 0:
        root.is_word = True
        return

    # split off the first character of the current word
    first_char = word[0:1]
    sufx = word[1:]

    # recursive step
    if first_char in root.children:
        add_to_trie(sufx, root.children[first_char])
    else:
        new_node = Node(first_char, False)
        root.children[first_char] = new_node
        add_to_trie(sufx, new_node)

def setup(words, root):
    for w in words:
        add_to_trie(w.strip(), root)

# convenience function to initiate recursion
def find_words(word, node):
    fwords = list()
    return find_words_r(word, node, fwords)

# recursive search
def find_words_r(word, node, fwords, path=""):

    # recursion base case
    if len(word) == 0:
        if node.is_word:
            fwords.append(path)
        return fwords

    # split the first character off the curent word
    first_char = word[0:1]
    sufx = word[1:] 

    # recursive step
    if first_char in node.children:
        return find_words_r(sufx, node.children[first_char], fwords, path + first_char)
    elif first_char == '.':
        for k in node.children.keys():
            find_words_r(sufx, node.children[k], fwords, path + k)
        return fwords
    elif first_char == '*':
        # if we encounter a '*' we exand it into '.' characters
        # we try every possible length of word 
        for k in range(1,MAX_ENG_WORD_LENGTH-(len(path)+ len(sufx))):
            expansion = "."*k
            sufx = expansion + sufx
            find_words_r(sufx, node, fwords, path)
        return fwords
    else:
        return fwords

# This function serializes a Trie Node into a string that can be stored. 
# This function is untested
def serialize(node):
    stk = list()
    serialized_trie = list()
    # load the stack with all children of root node
    stk.append(">")
    for k, v in node.children.items():
        stk.append(v)
    stk.append("<")

    while(len(stk)>0):
        cur_node = stk.pop()
        if not isinstance(cur_node, str):
            serialized_trie.append(cur_node.value)
            stk.append('>')
            for k, v in cur_node.children.items():
                stk.append(v)
            stk.append('<')
        else:
            serialized_trie.append(cur_node)

    return "".join(serialized_trie)

# This function builds a tree from a string representing a serialized
# tree. This function is untested. 
def deserialize(s):
    if not s or len(s) == 0:
        return None
    root = Node("",False)
    current = root
    stk = list()
    for c in s:
        if c == "<":
            stk.append(current)
        if c == ">":
            stk.pop()
        else:
            current = Node(c,"False")
            stk[-1].children[c] = current
    return root


import click
from lexgo import config
import subprocess
from lexgo import utils

# Global Constants
ENGLISH_DICT_PATH = "eng_words_alpha.txt"

@click.command()
@click.argument(
    "word"
)
@click.version_option()
@click.option("-e", "--exclude", default="",
    help="Individual letters that MUST NOT appear in found words.",
)
@click.option("-i", "--include", default="",
    help="Individual letters that MUST appear in found words.",
)
@click.option("-xp", type=(str, int), multiple=True, default=[],
    help="A letter and a position in which it must not appear.",
)
@click.option("-l", "--lang", type=click.Choice(['de', 'en', 'es', 'fr', 'pt'], case_sensitive=False), default='en',
    help="The language dictionary to search.",
)
def lexgo(word, exclude, include, xp, lang):
    '''
    Search for WORD.

    WORD can be made up of letters, dots ('.'), and stars ('*'). A dot is a placeholder for any
    one letter. A star is a placeholder for one or more letters.  

    EXAMPLES: 

    \b
    lexgo .est - search for words that start with any letter and end 'est'
    lexgo ..ed - search four letter words ending 'ed'
    lexgo *est - search all words that end in 'est'
    lexgo b.. -e td -i a -xp ns 3
               - search 3 letter words starting with b, without letters 't' or 'd',
                 with letter a, and without letters 'n' or 's' in the 3rd letter.
    '''
    # verify that grep is installed
    if not utils.is_grep_installed():
        raise click.FileError("grep", "A GNU compatible grep application must be installed to use lexgo.")
    
    # verify that word has only alpha '.' and '*'
    if word and not utils.is_alpha_dot_star(word): 
        raise click.UsageError("WORD must consist of alphabet characters, dot (.), and star (*).")
    
    if exclude and not utils.is_alpha(exclude):
        raise click.UsageError("OPTION --exclude only accepts alphabetic characters.")

    if include and not utils.is_alpha(include):
        raise click.UsageError("OPTION --include only accepts alphabetic characters.")

    if xp:
        for c, p in xp:
            if not utils.is_alpha(c): 
                raise click.UsageError("OPTION --xp only accepts alphabetic characters for first argument.")
            if p > config.LARGEST_WORD:
                raise click.UsageError("OPTION --xp only accepts integers less than 50 for second argument.")

    # convert the simple '*' into regular expression".*"
    word = str.replace(word, "*", ".*")
   
    # command stack
    command_stack = []

    # build initial grep command
    output = subprocess.Popen(["grep", "-w", "^" + word, config.DICT_PATHS[lang]], 
                        stdout=subprocess.PIPE, text=True)
    command_stack.append(output)

    # add exclusions
    if exclude:
        p = subprocess.Popen(["grep", "-v", "[{}]".format(exclude)], stdin=command_stack[-1].stdout, 
                                stdout=subprocess.PIPE, text=True)
        command_stack.append(p)
    
    # add inclusions
    if include:
        p = subprocess.Popen(["grep", "[{}]".format(include)], stdin=command_stack[-1].stdout, 
                                stdout=subprocess.PIPE, text=True)
        command_stack.append(p)

    # add positional exclusions (xp)
    if xp:
        for c, p in xp:
            gstr = "^" + "."*(p-1) + c
            p = subprocess.Popen(["grep", "-v", gstr], stdin=command_stack[-1].stdout, 
                                stdout=subprocess.PIPE, text=True)
            command_stack.append(p)

    # execute command stack
    out, error = command_stack[-1].communicate()
    if not error:
        click.echo(out)



import pathlib

# Global Constants
DATA_DIR_NAME = "data"
DICT_PATHS = {
    "en" : pathlib.Path(__file__).parent.joinpath(DATA_DIR_NAME, "en.txt").absolute().as_posix(),
    "es" : pathlib.Path(__file__).parent.joinpath(DATA_DIR_NAME, "es.txt").absolute().as_posix(),
    "fr" : pathlib.Path(__file__).parent.joinpath(DATA_DIR_NAME, "fr.txt").absolute().as_posix(),
    "pt" : pathlib.Path(__file__).parent.joinpath(DATA_DIR_NAME, "pt.txt").absolute().as_posix(),
    "de" : pathlib.Path(__file__).parent.joinpath(DATA_DIR_NAME, "de.txt").absolute().as_posix()
}

LARGEST_WORD = 50


import subprocess
import re

def is_grep_installed():
    try:
        cp = subprocess.run(["grep", "--version"], capture_output=True)
        return "grep" in cp.stdout.decode("utf-8")
    except FileNotFoundError:
        return False
    
def is_alpha():
    return True

def is_alpha_dot_star(str):
    return re.fullmatch(r"[\w\.\*]*",str) != None

def is_alpha(str):
    return re.fullmatch(r"[\w\.\*]*",str) != None


from click.testing import CliRunner
from lexgo.cli import lexgo


def test_version():
    runner = CliRunner()
    with runner.isolated_filesystem():
        result = runner.invoke(lexgo, ["--version"])
        assert result.exit_code == 0
        assert result.output.startswith("lexgo, version ")

def test_find_basic_word_eng():
    runner = CliRunner()
    with runner.isolated_filesystem():
        result = runner.invoke(lexgo, ["test"])
        assert result.exit_code == 0
        assert result.output.startswith("test")

def test_find_basic_word_es():
    runner = CliRunner()
    with runner.isolated_filesystem():
        result = runner.invoke(lexgo, ["-l", "es", "prueba"])
        assert result.exit_code == 0
        assert result.output.startswith("prueba")

def test_find_basic_word_fr():
    runner = CliRunner()
    with runner.isolated_filesystem():
        result = runner.invoke(lexgo, ["-l", "fr", "tester"])
        assert result.exit_code == 0
        assert result.output.startswith("tester")

def test_find_basic_word_de():
    runner = CliRunner()
    with runner.isolated_filesystem():
        result = runner.invoke(lexgo, ["-l", "de", "testen"])
        assert result.exit_code == 0
        assert result.output.startswith("testen")

''' Test falies with word "testar". I think it's an encoding issue I need to solve. 
def test_find_basic_word_pt():
    runner = CliRunner()
    with runner.isolated_filesystem():
        result = runner.invoke(lexgo, ["-l", "pt", "teste"])
        assert result.exit_code == 0
        assert result.output.startswith("teste")
'''